#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Original Version from:
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2015 Thomas Voegtlin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import Any, Dict, Tuple

from bitcoin_safe.fx import FX
from bitcoin_safe.gui.qt.wrappers import Menu

from ...config import UserConfig
from ...network_config import BlockchainType

logger = logging.getLogger(__name__)


import enum
from enum import IntEnum

import bdkpython as bdk
from PyQt6.QtCore import QModelIndex, QPoint, Qt, pyqtSignal
from PyQt6.QtGui import (
    QBrush,
    QColor,
    QDragEnterEvent,
    QDragMoveEvent,
    QDropEvent,
    QFont,
    QStandardItem,
)
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QComboBox,
    QHBoxLayout,
    QPushButton,
    QWidget,
)

from ...i18n import translate
from ...rpc import send_rpc_command
from ...signals import Signals, UpdateFilter, UpdateFilterReason, WalletSignals
from ...util import Satoshis, block_explorer_URL
from ...wallet import TxStatus, Wallet
from .category_list import CategoryEditor
from .my_treeview import (
    MyItemDataRole,
    MySortModel,
    MyStandardItemModel,
    MyTreeView,
    TreeViewWithToolbar,
    needs_frequent_flag,
)
from .taglist import AddressDragInfo
from .util import ColorScheme, Message, do_copy, read_QIcon, sort_id_to_icon, webopen


class ImportLabelMenu:
    def __init__(self, upper_menu: Menu, wallet: Wallet, wallet_signals: WalletSignals) -> None:
        self.wallet_signals = wallet_signals
        self.wallet = wallet
        self.import_label_menu = upper_menu.add_menu(
            "",
        )

        self.action_import = self.import_label_menu.add_action(
            "",
            lambda: self.wallet_signals.import_labels.emit(self.wallet.id),
        )
        self.action_bip329_import = self.import_label_menu.add_action(
            "",
            lambda: self.wallet_signals.import_bip329_labels.emit(self.wallet.id),
        )
        self.action_electrum_import = self.import_label_menu.add_action(
            "",
            lambda: self.wallet_signals.import_electrum_wallet_labels.emit(self.wallet.id),
        )
        self.action_nostr_import = self.import_label_menu.add_action(
            "",
            self.import_nostr_labels,
            icon=read_QIcon("cloud-sync.svg"),
        )
        self.updateUi()

    def import_nostr_labels(self):
        Message(
            translate(
                "import",
                "Please go to the Sync Tab and import your Sync key there. The labels will then be automatically restored.",
            )
        )

    def updateUi(self) -> None:
        self.import_label_menu.setTitle(translate("menu", "Import Labels"))
        self.action_import.setText(translate("menu", "Import Labels"))
        self.action_bip329_import.setText(translate("menu", "Import Labels (BIP329 / Sparrow)"))
        self.action_electrum_import.setText(translate("menu", "Import Labels (Electrum Wallet)"))
        self.action_nostr_import.setText(
            translate("menu", "Restore labels from cloud using an existing sync key")
        )


class ExportLabelMenu:
    def __init__(self, upper_menu: Menu, wallet: Wallet, wallet_signals: WalletSignals) -> None:
        self.wallet_signals = wallet_signals
        self.wallet = wallet
        self.export_label_menu = upper_menu.add_menu(
            "",
        )

        self.action_export_full = self.export_label_menu.add_action(
            "",
            lambda: self.wallet_signals.export_labels.emit(self.wallet.id),
        )
        self.action_bip329 = self.export_label_menu.add_action(
            "",
            lambda: self.wallet_signals.export_bip329_labels.emit(self.wallet.id),
        )
        self.updateUi()

    def updateUi(self) -> None:
        self.export_label_menu.setTitle(translate("menu", "Export Labels"))
        self.action_export_full.setText(translate("export", "Export Labels"))
        self.action_bip329.setText(translate("export", "Export Labels for other wallets (BIP329)"))


class AddressUsageStateFilter(IntEnum):
    ALL = 0
    UNUSED = 1
    FUNDED = 2
    USED_AND_EMPTY = 3
    FUNDED_OR_UNUSED = 4

    def ui_text(self) -> str:
        return {
            self.ALL: translate("address_list", "All status"),
            self.UNUSED: translate("address_list", "Unused"),
            self.FUNDED: translate("address_list", "Funded"),
            self.USED_AND_EMPTY: translate("address_list", "Used"),
            self.FUNDED_OR_UNUSED: translate("address_list", "Funded or Unused"),
        }[self]


class AddressTypeFilter(IntEnum):
    ALL = 0
    RECEIVING = 1
    CHANGE = 2

    def ui_text(self) -> str:
        return {
            self.ALL: translate("address_list", "All types"),
            self.RECEIVING: translate("address_list", "Receiving"),
            self.CHANGE: translate("address_list", "Change"),
        }[self]


class AddressList(MyTreeView):
    signal_tag_dropped = pyqtSignal(AddressDragInfo)

    class Columns(MyTreeView.BaseColumnsEnum):
        NUM_TXS = enum.auto()
        TYPE = enum.auto()
        INDEX = enum.auto()
        ADDRESS = enum.auto()
        CATEGORY = enum.auto()
        LABEL = enum.auto()
        COIN_BALANCE = enum.auto()
        FIAT_BALANCE = enum.auto()

    filter_columns = [
        Columns.TYPE,
        Columns.ADDRESS,
        Columns.CATEGORY,
        Columns.LABEL,
        Columns.COIN_BALANCE,
    ]
    column_alignments = {
        Columns.TYPE: Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter,
        Columns.ADDRESS: Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
        Columns.CATEGORY: Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter,
        Columns.LABEL: Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
        Columns.COIN_BALANCE: Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter,
        Columns.NUM_TXS: Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter,
        Columns.FIAT_BALANCE: Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter,
    }

    hidden_columns = [Columns.INDEX]

    stretch_column = Columns.LABEL
    key_column = Columns.ADDRESS
    column_widths: Dict[MyTreeView.BaseColumnsEnum, int] = {Columns.ADDRESS: 150, Columns.COIN_BALANCE: 100}

    def __init__(
        self,
        fx: FX,
        config: UserConfig,
        wallet: Wallet,
        wallet_signals: WalletSignals,
        signals: Signals,
    ) -> None:
        super().__init__(
            config=config,
            signals=signals,
            stretch_column=self.stretch_column,
            column_widths=self.column_widths,
            editable_columns=[AddressList.Columns.LABEL],
            sort_column=AddressList.Columns.COIN_BALANCE,
            sort_order=Qt.SortOrder.DescendingOrder,
        )
        self.fx = fx
        self.wallet_signals = wallet_signals
        self.wallet = wallet
        self.setTextElideMode(Qt.TextElideMode.ElideMiddle)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.show_change = AddressTypeFilter.ALL  # type: AddressTypeFilter
        self.show_used = AddressUsageStateFilter.ALL  # type: AddressUsageStateFilter
        self._source_model = MyStandardItemModel(self, drag_key="addresses")
        self.proxy = MySortModel(
            self, source_model=self._source_model, sort_role=MyItemDataRole.ROLE_SORT_ORDER
        )
        self.setModel(self.proxy)
        self.setSortingEnabled(True)  # Allow user to sort by clicking column headers
        self.updateUi()

        # signals
        self.wallet_signals.updated.connect(self.update_with_filter)
        self.wallet_signals.language_switch.connect(self.updateUi)
        self.fx.signal_data_updated.connect(self.on_update_fx_rates)

    def updateUi(self) -> None:
        self.update_content()

    def dragEnterEvent(self, event: QDragEnterEvent | None) -> None:
        super().dragEnterEvent(event)
        if not event or event.isAccepted():
            return

        if (mime_data := event.mimeData()) and self.get_json_mime_data(mime_data) is not None:
            event.acceptProposedAction()
        else:
            event.ignore()

    def dragMoveEvent(self, event: QDragMoveEvent | None) -> None:
        super().dragMoveEvent(event)
        if not event:
            return
        if event.isAccepted():
            return

        if (mime_data := event.mimeData()) and (
            json_mime_data := self.get_json_mime_data(mime_data)
        ) is not None:
            event.acceptProposedAction()
        else:
            event.ignore()

    def dropEvent(self, event: QDropEvent) -> None:  # type: ignore[override]
        # handle dropped files
        super().dropEvent(event)
        if event.isAccepted():
            return

        index = self.indexAt(event.position().toPoint())
        if not index.isValid():
            # Handle the case where the drop is not on a valid index
            return

        if (mime_data := event.mimeData()) and (
            json_mime_data := self.get_json_mime_data(mime_data)
        ) is not None:
            model = self.model()
            hit_address = model.data(model.index(index.row(), self.Columns.ADDRESS))
            if json_mime_data.get("type") == "drag_tag":
                if hit_address is not None:
                    drag_info = AddressDragInfo([json_mime_data.get("tag")], [hit_address])
                    logger.debug(f"drag_info {drag_info}")
                    self.signal_tag_dropped.emit(drag_info)
                event.accept()
                return

        event.ignore()

    def on_double_click(self, idx: QModelIndex) -> None:
        addr = self.get_role_data_for_current_item(col=self.key_column, role=MyItemDataRole.ROLE_KEY)
        self.wallet_signals.show_address.emit(addr, self.wallet.id)

    def get_address(self, force_new=False, category: str | None = None) -> bdk.AddressInfo:
        if force_new:
            address_info = self.wallet.get_address(force_new=force_new)
            address = address_info.address.as_string()
            self.wallet.labels.set_addr_category(address, category, timestamp="now")
            self.wallet_signals.updated.emit(
                UpdateFilter(addresses=set([address]), reason=UpdateFilterReason.NewAddressRevealed)
            )
        else:
            address_info = self.wallet.get_unused_category_address(category)
            address = address_info.address.as_string()

            if self.signals:
                self.signals.wallet_signals[self.wallet.id].updated.emit(
                    UpdateFilter(addresses=set([address]), reason=UpdateFilterReason.GetUnusedCategoryAddress)
                )

        do_copy(address, title=self.tr("Address {address}").format(address=address))
        self.select_row(address, self.Columns.ADDRESS)
        return address_info

    def toggle_change(self, state: int) -> None:
        if state == self.show_change:
            return
        self.show_change = AddressTypeFilter(state)
        self.update_content()

    def toggle_used(self, state: int) -> None:
        if state == self.show_used:
            return
        self.show_used = AddressUsageStateFilter(state)
        self.update_content()

    def on_update_fx_rates(self):
        addresses_with_balance = []

        model = self.sourceModel()
        for row in range(model.rowCount()):
            address = model.data(model.index(row, self.Columns.ADDRESS))
            balance = model.data(
                model.index(row, self.Columns.COIN_BALANCE), role=MyItemDataRole.ROLE_CLIPBOARD_DATA
            )
            if balance:
                addresses_with_balance.append(address)

        update_filter = UpdateFilter(addresses=addresses_with_balance, reason=UpdateFilterReason.NewFxRates)
        self.update_with_filter(update_filter)

    def update_with_filter(self, update_filter: UpdateFilter) -> None:
        if update_filter.refresh_all:
            return self.update_content()
        logger.debug(f"{self.__class__.__name__}  update_with_filter {update_filter}")

        self._before_update_content()
        remaining_addresses = set(update_filter.addresses)

        model = self.sourceModel()
        log_info = []
        # Select rows with an ID in id_list
        for row in range(model.rowCount()):
            address = model.data(model.index(row, self.Columns.ADDRESS))
            address_match = address in update_filter.addresses
            category_match = model.data(model.index(row, self.Columns.CATEGORY)) in update_filter.categories
            if (
                (
                    update_filter.reason == UpdateFilterReason.ChainHeightAdvanced
                    and model.data(
                        model.index(row, self.key_column), role=MyItemDataRole.ROLE_FREQUENT_UPDATEFLAG
                    )
                )
                or address_match
                or (not update_filter.addresses and category_match or len(update_filter.categories) > 1)
            ):
                log_info.append((row, address))
                self.refresh_row(address, row)
                remaining_addresses = remaining_addresses - set([address])

        # get_maximum_index
        # address_infos_min = max([self.wallet.get_address_info_min( address) for address in remaining_addresses ])
        # max_index =

        # sometimes additional addresses are updated,
        # i can add them here without recreating the whole model
        if remaining_addresses:
            for address in set(self.wallet.get_addresses()).intersection(remaining_addresses):
                log_info.append((0, address))
                self.append_address(address)
                remaining_addresses = remaining_addresses - set([address])

        logger.debug(f"Updated addresses  {log_info}.  remaining_addresses = {remaining_addresses}")
        self._after_update_content()

    def get_headers(self) -> Dict:
        return {
            self.Columns.NUM_TXS: self.tr("Tx"),
            self.Columns.TYPE: self.tr("Type"),
            self.Columns.INDEX: self.tr("Index"),
            self.Columns.ADDRESS: self.tr("Address"),
            self.Columns.CATEGORY: self.tr("Category"),
            self.Columns.LABEL: self.tr("Label"),
            self.Columns.COIN_BALANCE: self.tr("Balance"),
            self.Columns.FIAT_BALANCE: "$ " + self.tr("Balance"),
        }

    def update_content(self) -> None:
        if self.maybe_defer_update():
            return
        logger.debug(f"{self.__class__.__name__} update")
        self._before_update_content()

        current_selected_key = self.get_role_data_for_current_item(
            col=self.key_column, role=MyItemDataRole.ROLE_KEY
        )
        if self.show_change == AddressTypeFilter.RECEIVING:
            addr_list = self.wallet.get_receiving_addresses()
        elif self.show_change == AddressTypeFilter.CHANGE:
            addr_list = self.wallet.get_change_addresses()
        else:
            addr_list = self.wallet.get_addresses()
        self._source_model.clear()
        self.update_headers(self.get_headers())
        for address in addr_list:
            self.append_address(address)

        self._after_update_content()
        super().update_content()

    def append_address(self, address: str) -> None:
        balance = self.wallet.get_addr_balance(address).total
        is_used_and_empty = self.wallet.address_is_used(address) and balance == 0
        if self.show_used == AddressUsageStateFilter.UNUSED and (balance or is_used_and_empty):
            return
        if self.show_used == AddressUsageStateFilter.FUNDED and balance == 0:
            return
        if self.show_used == AddressUsageStateFilter.USED_AND_EMPTY and not is_used_and_empty:
            return
        if self.show_used == AddressUsageStateFilter.FUNDED_OR_UNUSED and is_used_and_empty:
            return
        labels = [""] * len(self.Columns)
        labels[self.Columns.ADDRESS] = address
        item = [QStandardItem(e) for e in labels]
        item[self.Columns.ADDRESS].setData(address, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        # align text and set fonts
        # for i, item in enumerate(item):
        #     item.setTextAlignment(Qt.AlignmentFlag.AlignVCenter)
        #     if i in (self.Columns.ADDRESS,):
        #         item.setFont(QFont(MONOSPACE_FONT))
        self.set_editability(item)
        # setup column 0

        address_info_min = self.wallet.get_address_info_min(address)
        if address_info_min:
            item[self.Columns.INDEX].setData(address_info_min.index, MyItemDataRole.ROLE_CLIPBOARD_DATA)
            if address_info_min.is_change():
                item[self.Columns.TYPE].setText(self.tr("change"))
                item[self.Columns.TYPE].setData(self.tr("change"), MyItemDataRole.ROLE_CLIPBOARD_DATA)
                item[self.Columns.TYPE].setBackground(ColorScheme.YELLOW.as_color(True))
            else:
                item[self.Columns.TYPE].setText(self.tr("receiving"))
                item[self.Columns.TYPE].setData(self.tr("receiving"), MyItemDataRole.ROLE_CLIPBOARD_DATA)
                item[self.Columns.TYPE].setBackground(ColorScheme.GREEN.as_color(True))
            item[self.key_column].setData(address, MyItemDataRole.ROLE_KEY)
            item[self.Columns.TYPE].setData(
                (address_info_min.address_path()[0], -address_info_min.address_path()[1]),
                MyItemDataRole.ROLE_SORT_ORDER,
            )
            item[self.Columns.TYPE].setToolTip(
                f"""{address_info_min.address_path()[1]}. {self.tr("change address") if address_info_min.address_path()[0] else   self.tr('receiving address')}"""
            )
        # add item
        count = self._source_model.rowCount()
        self._source_model.insertRow(count, item)
        self.refresh_row(address, count)

    def refresh_row(self, key: str, row: int) -> None:
        assert row is not None
        address = key
        label = self.wallet.get_label_for_address(address)
        category = self.wallet.labels.get_category(address)

        txids = self.wallet.get_involved_txids(address)
        fulltxdetails = [self.wallet.get_dict_fulltxdetail().get(txid) for txid in txids]
        txs_involed = [fulltxdetail.tx for fulltxdetail in fulltxdetails if fulltxdetail]

        statuses = [TxStatus.from_wallet(tx.txid, self.wallet) for tx in txs_involed]
        min_status = sorted(statuses, key=lambda status: status.sort_id())[0] if statuses else None
        icon_path = sort_id_to_icon(min_status.sort_id()) if min_status else None
        num = len(txs_involed)

        balance = self.wallet.get_addr_balance(address).total
        balance_text = str(Satoshis(balance, self.wallet.network))
        # create item

        dollar_amount = self.fx.to_fiat("usd", balance)
        fiat_balance_str = (
            self.fx.format_dollar(dollar_amount, prepend_dollar_sign=False)
            if dollar_amount is not None
            else ""
        )
        fiat_balance_data = (
            self.fx.format_dollar(dollar_amount, prepend_dollar_sign=False)
            if dollar_amount is not None
            else ""
        )
        _item = [self._source_model.item(row, col) for col in self.Columns]
        item = [entry for entry in _item if entry]
        if needs_frequent_flag(status=min_status):
            item[self.key_column].setData(True, role=MyItemDataRole.ROLE_FREQUENT_UPDATEFLAG)
        item[self.Columns.LABEL].setText(label)
        item[self.Columns.LABEL].setData(label, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        item[self.Columns.CATEGORY].setText(category if category else "")
        item[self.Columns.CATEGORY].setData(category, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        item[self.Columns.CATEGORY].setBackground(CategoryEditor.color(category))
        item[self.Columns.COIN_BALANCE].setText(balance_text)
        color = QColor(0, 0, 0) if balance else QColor(255 // 2, 255 // 2, 255 // 2)
        item[self.Columns.COIN_BALANCE].setForeground(QBrush(color))
        item[self.Columns.COIN_BALANCE].setData(balance, MyItemDataRole.ROLE_SORT_ORDER)
        item[self.Columns.COIN_BALANCE].setData(balance, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        item[self.Columns.FIAT_BALANCE].setText(fiat_balance_str)
        item[self.Columns.FIAT_BALANCE].setData(fiat_balance_data, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        # item[self.Columns.NUM_TXS].setText("%d" % num)
        item[self.Columns.NUM_TXS].setToolTip(f"{num} Transaction")
        item[self.Columns.NUM_TXS].setData(num, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        item[self.Columns.NUM_TXS].setIcon(read_QIcon(icon_path))

        # calculated_width = QFontMetrics(self.font()).horizontalAdvance(balance_text)
        # current_width = self.header().sectionSize(self.Columns.ADDRESS)
        # # Update the column width if the calculated width is larger
        # if calculated_width > current_width:
        #     self.header().resizeSection(self.Columns.ADDRESS, calculated_width)

    def create_menu(self, position: QPoint) -> Menu:
        menu = Menu()
        # is_multisig = isinstance(self.wallet, Multisig_Wallet)
        selected = self.selected_in_column(self.Columns.ADDRESS)
        if not selected:
            return menu
        multi_select = len(selected) > 1
        selected_items = [self.item_from_index(item) for item in selected]
        addrs = [item.text() for item in selected_items if item]
        if not multi_select:
            idx = self.indexAt(position)
            if not idx.isValid():
                return menu
            item = self.item_from_index(idx)
            if not item:
                return menu
            addr = addrs[0]
            menu.add_action(
                self.tr("Details"), lambda: self.wallet_signals.show_address.emit(addr, self.wallet.id)
            )

            addr_URL = block_explorer_URL(self.config.network_config.mempool_url, "addr", addr)
            if addr_URL:
                menu.add_action(
                    self.tr("View on block explorer"), lambda: webopen(addr_URL), icon=read_QIcon("link.svg")
                )

            menu.addSeparator()

            self.add_copy_menu(menu, idx, include_columns_even_if_hidden=[self.key_column])

        menu.add_action(
            self.tr("Copy as csv"),
            lambda: self.copyRowsToClipboardAsCSV([r.row() for r in selected]),
            icon=read_QIcon("csv-file.svg"),
        )
        menu.addSeparator()
        self.export_label_menu = ExportLabelMenu(menu, wallet=self.wallet, wallet_signals=self.wallet_signals)
        self.import_label_menu = ImportLabelMenu(menu, wallet=self.wallet, wallet_signals=self.wallet_signals)

        # run_hook('receive_menu', menu, addrs, self.wallet)
        if viewport := self.viewport():
            menu.exec(viewport.mapToGlobal(position))

        return menu

    # def place_text_on_clipboard(self, text: str, *, title: str = None) -> None:
    #     if bdk.Address(text):
    #         try:
    #             self.wallet.check_address_for_corruption(text)
    #         except InternalAddressCorruption as e:
    #             self.show_error(str(e))
    #             raise
    #     super().place_text_on_clipboard(text, title=title)

    def get_edit_key_from_coordinate(self, row, col) -> Any:
        if col != self.Columns.LABEL:
            return None
        return self.get_role_data_from_coordinate(row, self.key_column, role=MyItemDataRole.ROLE_KEY)

    def on_edited(self, idx, edit_key, *, text) -> None:
        self.wallet.labels.set_addr_label(edit_key, text, timestamp="now")
        categories = []
        if not self.wallet.labels.get_category_raw(edit_key):
            # also fix the category to have consitency across wallets via the labelsyncer
            category = self.wallet.labels.get_category(edit_key)
            categories += [category]
            self.wallet.labels.set_addr_category(edit_key, category, timestamp="now")
        self.wallet_signals.updated.emit(
            UpdateFilter(
                addresses=[edit_key],
                categories=categories,
                txids=self.wallet.get_involved_txids(edit_key),
                reason=UpdateFilterReason.UserInput,
            )
        )


class AddressListWithToolbar(TreeViewWithToolbar):
    def __init__(
        self,
        address_list: AddressList,
        config: UserConfig,
        parent: QWidget | None = None,
        signals: Signals | None = None,
    ) -> None:
        super().__init__(address_list, config, parent=parent)
        self.signals = signals
        self.address_list: AddressList = address_list
        self.change_button = QComboBox(self)
        self.change_button.currentIndexChanged.connect(self.address_list.toggle_change)
        for addr_type in AddressTypeFilter.__members__.values():
            self.change_button.addItem(addr_type.ui_text())
        self.used_button = QComboBox(self)
        self.used_button.currentIndexChanged.connect(self.address_list.toggle_used)
        for addr_usage_state in AddressUsageStateFilter.__members__.values():
            self.used_button.addItem(addr_usage_state.ui_text())

        self.create_layout()
        self.updateUi()

        self.address_list.wallet_signals.language_switch.connect(self.updateUi)
        self.address_list.wallet_signals.updated.connect(self.updateUi)

    def updateUi(self) -> None:
        super().updateUi()

        self.action_show_filter.setText(self.tr("Show Filter"))
        self.menu_import_labels.updateUi()
        self.menu_export_labels.updateUi()

        if self.balance_label:
            balance = self.address_list.wallet.get_balance()
            if self.signals:
                display_balance = (
                    self.signals.wallet_signals[self.address_list.wallet.id]
                    .get_display_balance.emit()
                    .get(self.address_list.wallet.id)
                )
                if display_balance:
                    balance = display_balance

            self.balance_label.setText(balance.format_short(self.address_list.wallet.network))
            self.balance_label.setToolTip(balance.format_long(self.address_list.wallet.network))

    def create_toolbar_with_menu(self, title) -> None:
        super().create_toolbar_with_menu(title=title)

        font = QFont()
        font.setPointSize(12)
        self.balance_label.setFont(font)

        self.action_show_filter = self.menu.addToggle("", lambda: self.toggle_toolbar(self.config))
        self.menu_export_labels = ExportLabelMenu(
            self.menu, wallet=self.address_list.wallet, wallet_signals=self.address_list.wallet_signals
        )
        self.menu_import_labels = ImportLabelMenu(
            self.menu, wallet=self.address_list.wallet, wallet_signals=self.address_list.wallet_signals
        )

        if (
            self.config
            and self.config.network_config.server_type == BlockchainType.RPC
            and self.config.network != bdk.Network.BITCOIN
        ):

            def mine_to_selected_addresses() -> None:
                selected = self.address_list.selected_in_column(self.address_list.Columns.ADDRESS)
                if not selected:
                    return
                selected_items = [self.address_list.item_from_index(item) for item in selected]
                addresses = [item.text() for item in selected_items if item]

                for address in addresses:
                    response = send_rpc_command(
                        self.config.network_config.rpc_ip,
                        str(self.config.network_config.rpc_port),
                        self.config.network_config.rpc_username,
                        self.config.network_config.rpc_password,
                        "generatetoaddress",
                        params=[1, address],
                    )
                    logger.info(f"{response}")
                if self.signals:
                    self.signals.chain_data_changed.emit(f"Mined to addresses {addresses}")

            b = QPushButton(self.tr("Generate to selected adddresses"))
            b.clicked.connect(mine_to_selected_addresses)
            self.toolbar.insertWidget(self.toolbar.count() - 2, b)

        hbox = self.create_toolbar_buttons()
        self.toolbar.insertLayout(self.toolbar.count() - 1, hbox)

    def create_toolbar_buttons(self) -> QHBoxLayout:
        def get_toolbar_buttons() -> Tuple[QComboBox, QComboBox]:
            return self.change_button, self.used_button

        hbox = QHBoxLayout()
        buttons = get_toolbar_buttons()
        for b in buttons:
            b.setVisible(False)
            hbox.addWidget(b)
        self.toolbar_buttons = buttons
        return hbox

    def on_hide_toolbar(self) -> None:
        self.update()

    def show_toolbar(self, is_visible: bool, config=None) -> None:
        super().show_toolbar(is_visible=is_visible, config=config)
        for b in self.toolbar_buttons:
            b.setVisible(is_visible)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import List

from bitcoin_safe.gui.qt.dialogs import question_dialog
from bitcoin_safe.gui.qt.util import Message, MessageType, caught_exception_message
from bitcoin_safe.i18n import translate
from bitcoin_safe.psbt_util import PubKeyInfo

from .dynamic_lib_load import setup_libsecp256k1
from .gui.qt.dialog_import import ImportDialog

setup_libsecp256k1()


logger = logging.getLogger(__name__)

import bdkpython as bdk
from bitcoin_qr_tools.bitcoin_video_widget import BitcoinVideoWidget
from bitcoin_qr_tools.data import Data, DataType
from bitcoin_usb.gui import USBGui
from bitcoin_usb.psbt_tools import PSBTTools
from bitcoin_usb.software_signer import SoftwareSigner
from PyQt6.QtCore import QObject, pyqtSignal

from .keystore import KeyStoreImporterTypes
from .util import tx_of_psbt_to_hex, tx_to_hex
from .wallet import Wallet


class AbstractSignatureImporter(QObject):
    signal_signature_added = pyqtSignal(bdk.PartiallySignedTransaction)
    signal_final_tx_received = pyqtSignal(bdk.Transaction)
    keystore_type = KeyStoreImporterTypes.watch_only

    def __init__(
        self,
        network: bdk.Network,
        signature_available: bool = False,
        key_label: str = "",
        pub_keys_without_signature: List[PubKeyInfo] | None = None,
    ) -> None:
        super().__init__()
        self.network = network
        self.signature_available = signature_available
        self.key_label = key_label
        self.pub_keys_without_signature = pub_keys_without_signature

    def sign(self, psbt: bdk.PartiallySignedTransaction, sign_options: bdk.SignOptions | None = None):
        pass

    @property
    def label(self) -> str:
        return ""

    def get_singing_options(self):
        pass

    def can_sign(self) -> bool:
        return False

    def txids_match(
        self,
        psbt1: bdk.PartiallySignedTransaction,
        psbt2: bdk.PartiallySignedTransaction,
    ) -> bool:
        return bool(psbt1.txid() == psbt2.txid())


class SignatureImporterWallet(AbstractSignatureImporter):
    keystore_type = KeyStoreImporterTypes.seed

    def __init__(
        self, wallet: Wallet, network: bdk.Network, signature_available: bool = False, key_label: str = ""
    ) -> None:
        super().__init__(
            network=network,
            signature_available=signature_available,
            key_label=key_label,
            pub_keys_without_signature=[
                PubKeyInfo(keystore.fingerprint, label=keystore.label)
                for keystore in wallet.keystores
                if keystore.mnemonic
            ],
        )

        self.software_signers = [
            SoftwareSigner(keystore.mnemonic, self.network)
            for keystore in wallet.keystores
            if keystore.mnemonic
        ]

    def can_sign(self) -> bool:
        return bool(self.software_signers)

    def sign(self, psbt: bdk.PartiallySignedTransaction, sign_options: bdk.SignOptions | None = None):
        original_psbt = psbt
        original_serialized_tx = tx_of_psbt_to_hex(psbt)
        for software_signer in self.software_signers:
            psbt = software_signer.sign_psbt(psbt)

        if not self.txids_match(original_psbt, psbt):
            Message(self.tr("The txid of the signed psbt doesnt match the original txid. Aborting"))
            return

        logger.debug(f"psbt before signing: {tx_of_psbt_to_hex(psbt)}")

        signing_was_successful: bool = original_serialized_tx != tx_of_psbt_to_hex(psbt)

        if signing_was_successful:
            logger.debug(f"psbt after signing: {tx_of_psbt_to_hex(psbt)}")
            logger.debug(f"psbt after signing: fee  {psbt.fee_rate().as_sat_per_vb()}")

        else:
            logger.debug(f"signign not completed")
        self.signal_signature_added.emit(psbt)

    @property
    def label(self) -> str:
        return self.tr("Sign with mnemonic seed")


class SignatureImporterQR(AbstractSignatureImporter):
    keystore_type = KeyStoreImporterTypes.qr

    def __init__(
        self,
        network: bdk.Network,
        signature_available: bool = False,
        key_label: str = "",
        pub_keys_without_signature=None,
        label: str | None = None,
    ) -> None:
        super().__init__(
            network=network,
            signature_available=signature_available,
            key_label=key_label,
            pub_keys_without_signature=pub_keys_without_signature,
        )
        self._label = label if label else self.tr("Scan QR code")
        self._temp_bitcoin_video_widget: BitcoinVideoWidget | None = None

    def scan_result_callback(self, original_psbt: bdk.PartiallySignedTransaction, data: Data):
        logger.debug(str(data.data))
        if data.data_type == DataType.PSBT:
            scanned_psbt: bdk.PartiallySignedTransaction = data.data

            if not self.txids_match(scanned_psbt, original_psbt):
                Message(
                    self.tr("The txid of the signed psbt doesnt match the original txid"),
                    type=MessageType.Error,
                )
                return

            logger.debug(str(scanned_psbt.serialize()))
            psbt2 = original_psbt.combine(scanned_psbt)

            if not self.txids_match(psbt2, original_psbt):
                Message(
                    self.tr("The txid of the signed psbt doesnt match the original txid"),
                    type=MessageType.Error,
                )
                return

            if psbt2.serialize() == original_psbt.serialize():
                Message(
                    self.tr("No additional signatures were added"),
                    type=MessageType.Error,
                )
                return

            # check if the tx can be finalized:
            finalized_tx = PSBTTools.finalize(psbt2, network=self.network)
            if finalized_tx:
                assert finalized_tx.txid() == original_psbt.txid(), self.tr(
                    "bitcoin_tx libary error. The txid should not be changed during finalizing"
                )
                self.signal_final_tx_received.emit(finalized_tx)
                return

            logger.debug(f"psbt updated {psbt2.serialize()}")
            self.signal_signature_added.emit(psbt2)

        elif data.data_type == DataType.Tx:
            scanned_tx: bdk.Transaction = data.data
            if scanned_tx.txid() != original_psbt.txid():
                Message(
                    self.tr("The txid of the signed psbt doesnt match the original txid"),
                    type=MessageType.Error,
                )
                return
            if tx_to_hex(scanned_tx) == tx_to_hex(original_psbt.extract_tx()):
                Message(
                    self.tr("No additional signatures were added"),
                    type=MessageType.Error,
                )
                return

            # TODO: Actually check if the tx is fully signed
            self.signal_final_tx_received.emit(scanned_tx)
        else:
            logger.warning(f"Datatype {data.data_type} is not valid for importing signatures")

    def sign(self, psbt: bdk.PartiallySignedTransaction, sign_options: bdk.SignOptions | None = None):
        if self._temp_bitcoin_video_widget:
            self._temp_bitcoin_video_widget.close()
        self._temp_bitcoin_video_widget = BitcoinVideoWidget(network=self.network)
        self._temp_bitcoin_video_widget.signal_data.connect(
            lambda data: self.scan_result_callback(psbt, data)
        )
        self._temp_bitcoin_video_widget.show()

    @property
    def label(self) -> str:
        return f"{self._label}"


class SignatureImporterFile(SignatureImporterQR):
    keystore_type = KeyStoreImporterTypes.file

    def __init__(
        self,
        network: bdk.Network,
        signature_available: bool = False,
        key_label: str = "",
        pub_keys_without_signature=None,
        label: str = translate("importer", "Import file"),
    ) -> None:
        super().__init__(
            network=network,
            signature_available=signature_available,
            key_label=key_label,
            pub_keys_without_signature=pub_keys_without_signature,
            label=label,
        )

    def sign(self, psbt: bdk.PartiallySignedTransaction, sign_options: bdk.SignOptions | None = None):
        tx_dialog = ImportDialog(
            network=self.network,
            window_title=self.tr("Import signed PSBT"),
            on_open=lambda s: self.scan_result_callback(psbt, Data.from_str(s, network=self.network)),
            text_button_ok=self.tr("OK"),
            text_instruction_label=self.tr("Please paste your PSBT in here, or drop a file"),
            text_placeholder=self.tr("Paste your PSBT in here or drop a file"),
        )
        tx_dialog.show()
        # tx_dialog.text_edit.button_open_file.click()

    @property
    def label(self) -> str:
        return f"{self._label}"


class SignatureImporterClipboard(SignatureImporterFile):
    keystore_type = KeyStoreImporterTypes.clipboard

    def __init__(
        self,
        network: bdk.Network,
        signature_available: bool = False,
        key_label: str = "",
        pub_keys_without_signature=None,
        label: str = translate("importer", "Import Signature"),
    ) -> None:
        super().__init__(
            network=network,
            signature_available=signature_available,
            key_label=key_label,
            pub_keys_without_signature=pub_keys_without_signature,
            label=label,
        )

    def sign(self, psbt: bdk.PartiallySignedTransaction, sign_options: bdk.SignOptions | None = None):
        tx_dialog = ImportDialog(
            network=self.network,
            window_title=self.tr("Import signed PSBT"),
            on_open=lambda s: self.scan_result_callback(psbt, Data.from_str(s, network=self.network)),
            text_button_ok=self.tr("OK"),
            text_instruction_label=self.tr("Please paste your PSBT in here, or drop a file"),
            text_placeholder=self.tr("Paste your PSBT in here or drop a file"),
        )
        tx_dialog.show()

    @property
    def label(self) -> str:
        return f"{self._label}"


class SignatureImporterUSB(SignatureImporterQR):
    keystore_type = KeyStoreImporterTypes.hwi

    def __init__(
        self,
        network: bdk.Network,
        signature_available: bool = False,
        key_label: str = "",
        pub_keys_without_signature=None,
        label: str | None = None,
    ) -> None:
        label = label if label else self.tr("USB Signing")
        super().__init__(
            network=network,
            signature_available=signature_available,
            key_label=key_label,
            pub_keys_without_signature=pub_keys_without_signature,
            label=label,
        )
        self.usb = USBGui(self.network)

    def sign(self, psbt: bdk.PartiallySignedTransaction, sign_options: bdk.SignOptions | None = None):
        try:
            signed_psbt = self.usb.sign(psbt)
            if signed_psbt:
                self.scan_result_callback(psbt, Data.from_psbt(signed_psbt))
        except Exception as e:
            if "multisig" in str(e).lower():
                question_dialog(
                    self.tr(
                        "Please do 'Wallet --> Export --> Export for ...' and register the multisignature wallet on the hardware signer."
                    )
                )
            else:
                caught_exception_message(e)

    @property
    def label(self) -> str:
        return f"{self._label}"



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from bitcoin_safe.gui.qt.data_tab_widget import DataTabWidget
from bitcoin_safe.html import html_f
from bitcoin_safe.signals import Signals

logger = logging.getLogger(__name__)

from bdkpython import Network
from PyQt6.QtCore import QObject, Qt, pyqtSignal
from PyQt6.QtSvgWidgets import QSvgWidget
from PyQt6.QtWidgets import (
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from ...util import call_call_functions
from .util import add_centered_icons, icon_path, qresize, read_QIcon


class NewWalletWelcomeScreen(QObject):
    signal_onclick_multisig_signature = pyqtSignal()
    signal_onclick_single_signature = pyqtSignal()
    signal_onclick_custom_signature = pyqtSignal()

    def __init__(self, main_tabs: DataTabWidget[object], network: Network, signals: Signals) -> None:
        super().__init__()
        self.main_tabs = main_tabs
        self.signals = signals

        self.name = "New wallet tab"
        self.network = network

        self.create_ui()

        self.pushButton_multisig.clicked.connect(
            lambda: call_call_functions([self.signal_onclick_multisig_signature.emit, self.remove_tab])
        )
        self.pushButton_singlesig.clicked.connect(
            lambda: call_call_functions([self.signal_onclick_single_signature.emit, self.remove_tab])
        )
        self.pushButton_custom_wallet.clicked.connect(
            lambda: call_call_functions([self.signal_onclick_custom_signature.emit, self.remove_tab])
        )
        logger.debug(f"initialized welcome_screen = {self}")

    def add_new_wallet_welcome_tab(self) -> None:
        self.main_tabs.add_tab(
            tab=self.tab,
            icon=read_QIcon("file.png"),
            description=self.tr("Create new wallet"),
            focus=True,
            data=self,
        )

    def create_ui(self) -> None:
        self.tab = QWidget()
        self.horizontalLayout_2 = QHBoxLayout(self.tab)
        self.groupBox_singlesig = QGroupBox(self.tab)
        self.verticalLayout = QVBoxLayout(self.groupBox_singlesig)
        self.label_singlesig = QLabel(self.groupBox_singlesig)
        # font = QFont()
        # font.setPointSize(11)
        # self.label_singlesig.setFont(font)
        self.label_singlesig.setWordWrap(True)

        self.verticalLayout.addWidget(self.label_singlesig)

        self.groupBox_1signingdevice = QGroupBox(self.groupBox_singlesig)
        self.groupBox_1signingdevice.setEnabled(True)
        self.horizontalLayout_4 = QHBoxLayout(self.groupBox_1signingdevice)

        self.svg_widget = QSvgWidget(icon_path("coldcard-only.svg"))
        self.svg_widget.setMinimumSize(qresize(self.svg_widget.sizeHint(), (60, 80)))
        self.svg_widget.setMaximumSize(qresize(self.svg_widget.sizeHint(), (60, 80)))
        self.horizontalLayout_4.addWidget(self.svg_widget)

        # set size of groupbox according to svg
        self.groupBox_1signingdevice.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

        self.verticalLayout.addWidget(self.groupBox_1signingdevice)

        self.pushButton_singlesig = QPushButton(self.groupBox_singlesig)

        self.verticalLayout.addWidget(self.pushButton_singlesig)

        self.horizontalLayout_2.addWidget(self.groupBox_singlesig)

        self.groupBox_multisig = QGroupBox(self.tab)
        self.verticalLayout_multisig = QVBoxLayout(self.groupBox_multisig)
        self.label_multisig = QLabel(self.groupBox_multisig)
        # font1 = QFont()
        # font1.setFamily(u"Noto Sans")
        # # font1.setPointSize(11)
        # font1.setBold(False)
        # font1.setItalic(False)
        # font1.setWeight(50)
        # self.label_multisig.setFont(font1)
        self.label_multisig.setWordWrap(True)

        self.verticalLayout_multisig.addWidget(self.label_multisig)

        self.groupBox_3signingdevices = QGroupBox(self.groupBox_multisig)
        self.groupBox_3signingdevices.setEnabled(True)
        self.groupBox_3signingdevices_layout = QHBoxLayout(self.groupBox_3signingdevices)

        add_centered_icons(
            ["coldcard-only.svg"] * 2 + ["bitbox02.svg"],
            self.groupBox_3signingdevices_layout,
            max_sizes=[(60, 80), (60, 80), (60, 50)],
        )

        self.verticalLayout_multisig.addWidget(self.groupBox_3signingdevices)
        # set size of groupbox according to svg
        self.groupBox_3signingdevices.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

        self.pushButton_multisig = QPushButton(self.groupBox_multisig)

        self.verticalLayout_multisig.addWidget(self.pushButton_multisig)

        self.horizontalLayout_2.addWidget(self.groupBox_multisig)

        self.groupBox_3 = QGroupBox(self.tab)
        self.verticalLayout_2 = QVBoxLayout(self.groupBox_3)
        self.label_custom = QLabel(self.groupBox_3)
        # self.label_custom.setFont(font)
        self.label_custom.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.label_custom)

        self.pushButton_custom_wallet = QPushButton(self.groupBox_3)

        self.verticalLayout_2.addWidget(self.pushButton_custom_wallet)

        self.horizontalLayout_2.addWidget(self.groupBox_3)

        self.label_singlesig.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.label_multisig.setAlignment(Qt.AlignmentFlag.AlignTop)
        # self.groupBox_3.setTitle(QCoreApplication.translate("Form", u"Custom", None))
        self.label_custom.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.updateUi()
        self.signals.language_switch.connect(self.updateUi)

    def updateUi(self) -> None:
        self.label_singlesig.setText(
            f"""<h1>{self.tr('Single Signature Wallet')}</h1>
<ul>
<li>{self.tr('Best for medium-sized funds')}</li>
</ul>             
<p><b>{self.tr('Pros:')}</b></p>
<ul>
<li>{self.tr('1 seed (24 secret words) is all you need to access your funds')}</li>
<li>{self.tr('1 secure location to store the seed backup (on paper or steel) is needed')}</li>
</ul>
<p><b>{self.tr('Cons:')}</b></p>
<ul>
<li>{self.tr('If you get tricked into giving hackers your seed, your Bitcoin will be stolen immediately')}</li>
</ul>""",
        )
        self.groupBox_1signingdevice.setTitle(self.tr("1 signing devices"))
        self.pushButton_singlesig.setText(self.tr("Choose Single Signature"))
        self.label_multisig.setText(
            f"""<h1>{self.tr('2 of 3 Multi-Signature Wal')}let</h1>
<ul>
<li>{self.tr('Best for large funds')}</li>
</ul>             
<p><b>{self.tr('Pros:')}</b></p>
<ul>
<li>{self.tr('If 1 seed was lost or stolen, all the funds can be transferred to a new wallet with the 2 remaining seeds + wallet descriptor (QR-code)')}</li>
</ul>
<p><b>{self.tr('Cons:')}</b></p>
<ul>
<li>{self.tr('3 secure locations (each with 1 seed backup   + wallet descriptor   are needed)')}</li>
<li>{self.tr('The wallet descriptor (QR-code) is necessary to recover the wallet')}</li>
</ul>
""",
        )
        self.groupBox_3signingdevices.setTitle(self.tr("3 signing devices"))
        self.pushButton_multisig.setText(self.tr("Choose Multi-Signature"))
        self.label_custom.setText(
            html_f(
                f"""<h1>{self.tr('Custom or restore existing Wallet')}</h1>
                <p><b>{self.tr('Pros:')}</b></p><p>{self.tr('Customize the wallet to your needs')}</p>
                <p><b>{self.tr('Cons:')}</b></p><p>{self.tr('Less support material online in case of recovery')}</p>""",
                add_html_and_body=True,
            )
        )
        self.pushButton_custom_wallet.setText(self.tr("Create custom wallet"))

    def remove_tab(self) -> None:
        index = self.main_tabs.indexOf(self.tab)
        if index >= 0 and self.main_tabs.count() > 1:
            self.main_tabs.removeTab(index)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import os
import platform
import shutil
import tempfile
from pathlib import Path
from typing import Any, List, Optional

from packaging import version
from PyQt6.QtWidgets import QHBoxLayout, QStyle, QWidget

from bitcoin_safe.gui.qt.downloader import Downloader, DownloadThread
from bitcoin_safe.gui.qt.notification_bar import NotificationBar
from bitcoin_safe.threading_manager import TaskThread, ThreadingManager

from ... import __version__
from ...html import html_f
from ...signals import SignalsMin
from ...signature_manager import (
    Asset,
    GitHubAssetDownloader,
    KnownGPGKeys,
    SignatureVerifyer,
)
from .util import Message, MessageType

logger = logging.getLogger(__name__)


class UpdateNotificationBar(NotificationBar, ThreadingManager):
    key = KnownGPGKeys.andreasgriffin

    def __init__(
        self, signals_min: SignalsMin, parent=None, threading_parent: ThreadingManager | None = None
    ) -> None:
        self.download_container = QWidget()
        self.download_container_layout = QHBoxLayout(self.download_container)
        current_margins = self.download_container_layout.contentsMargins()
        self.download_container_layout.setContentsMargins(
            current_margins.left(), 1, current_margins.right(), 0
        )  # Left, Top, Right, Bottom margins
        self.download_container_layout.setSpacing(self.download_container_layout.spacing() // 2)

        super().__init__(
            text="",
            optional_button_text="",
            callback_optional_button=lambda: self.check(),
            additional_widget=self.download_container,
            has_close_button=True,
            parent=parent,
            threading_parent=threading_parent,
        )
        self.signals_min = signals_min
        refresh_icon = (self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_BrowserReload)
        self.optionalButton.setIcon(refresh_icon)

        self.verifyer = SignatureVerifyer(list_of_known_keys=[self.key])
        self.assets: List[Asset] = []
        self.setVisible(False)

        self.refresh()
        self.signals_min.language_switch.connect(self.refresh)

    def get_asset_tag(self) -> Optional[str]:
        if self.assets:
            tag = self.assets[-1].tag
            return tag
        else:
            return None

    def is_new_version_available(self) -> bool:
        tag = self.get_asset_tag()
        if not tag:
            return False
        return version.parse(tag) > version.parse(__version__)

    def refresh(self) -> None:
        self.optionalButton.setText(self.tr("Check for Update"))

        # clear layout
        while self.download_container_layout.count():
            if (layout_item := self.download_container_layout.takeAt(0)) and (
                _widget := layout_item.widget()
            ):
                _widget.deleteLater()

        self.download_container.setVisible(bool(self.assets))
        if self.assets:
            if self.is_new_version_available():
                self.textLabel.setText(
                    self.tr("New version available {tag}").format(tag=self.get_asset_tag())
                )
                self.optionalButton.setVisible(False)
                self.setVisible(True)

                for asset in self.assets:
                    downloader = Downloader(url=asset.url, destination_dir=tempfile.gettempdir())
                    downloader.finished.connect(self.on_download_finished)
                    self.download_container_layout.addWidget(downloader)
            else:
                self.textLabel.setText(self.tr("You have already the newest version."))
                self.optionalButton.setVisible(True)
        else:
            self.textLabel.setText(self.tr("No update found"))
            self.optionalButton.setVisible(True)

    def on_download_finished(self, download_thread: DownloadThread) -> None:
        sig_file_path = self.verifyer.get_signature_from_web(download_thread.filename)
        if not sig_file_path:
            Message(self.tr("Could not verify the download. Please try again later."))
            self.refresh()
            self.setVisible(False)
            return

        destination = self.get_download_folder()
        was_signature_verified = None

        was_signature_verified = self.verifyer.verify_signature(
            download_thread.filename, expected_public_key=self.key
        )
        if not was_signature_verified:
            Message(self.tr("Signature doesn't match!!! Please try again."), type=MessageType.Error)
            self.refresh()
            self.setVisible(False)
            return

        self.textLabel.setText(html_f(self.tr("Signature verified."), color="green", bf=True))

        # overwrite the download_thread.filename so the show-button still works
        download_thread.filename = self.move_and_overwrite(download_thread.filename, destination)
        if sig_file_path:
            self.move_and_overwrite(sig_file_path, destination)

    @staticmethod
    def move_and_overwrite(source: Path, destination: Path) -> Path:
        # convert destination to destination with filename
        if destination.is_dir():
            destination = destination / source.name

        if os.path.exists(destination):
            os.remove(destination)
        shutil.move(source, destination)
        return destination

    @staticmethod
    def get_download_folder() -> Path:
        return Path.home() / "Downloads"

    def get_filtered_assets(self, assets: List[Asset]) -> List[Asset]:
        filtered_assets: List[Asset] = []
        for asset in assets:
            if platform.system() == "Windows" and not any(
                [asset.name.endswith(ending) for ending in ["exe", "msi"]]
            ):
                continue
            elif platform.system() == "Linux" and not any(
                [
                    asset.name.endswith(ending)
                    for ending in ["AppImage", "deb", "rpm", "flatpak", "snap", "pkg.tar.zst"]
                ]
            ):
                continue
            elif platform.system() == "Darwin" and not any(
                [asset.name.endswith(ending) for ending in ["dmg"]]
            ):
                continue

            # check if the asset can be recognized
            if not self.key.get_tag_if_mine(asset.name):
                continue

            filtered_assets.append(asset)
        return filtered_assets

    def check(self) -> None:
        def do() -> Any:
            return GitHubAssetDownloader(self.key.repository).get_assets_latest()

        def on_done(result) -> None:
            pass

        def on_success(assets: List[Asset]) -> None:
            # filter the assets, by recognized and for the platform

            self.assets = self.get_filtered_assets(assets)
            self.refresh()

        def on_error(packed_error_info) -> None:
            logger.error(f"error in fetching update info {packed_error_info}")

        self.append_thread(TaskThread().add_and_start(do, on_success, on_done, on_error))

    def check_and_make_visible(self) -> None:
        self.check()
        self.setVisible(True)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import Any, Callable, Optional, Union

from PyQt6.QtCore import pyqtBoundSignal
from PyQt6.QtGui import QAction, QIcon
from PyQt6.QtWidgets import QMenu, QMenuBar


class Menu(QMenu):
    def add_action(
        self,
        text="",
        slot: Optional[Union[Callable[[], Any], pyqtBoundSignal]] = None,
        icon: QIcon | None = None,
    ) -> QAction:
        action = QAction(text=text, parent=self)
        if slot:
            if callable(slot):
                action.triggered.connect(lambda: slot())
            else:
                action.triggered.connect(slot)
        self.addAction(action)
        if icon:
            action.setIcon(icon)
        return action

    def add_menu(self, text="") -> "Menu":
        menu = Menu(text, self)
        self.addMenu(menu)
        return menu


class MenuBar(QMenuBar):
    def add_menu(self, text="") -> Menu:
        menu = Menu(text, self)
        self.addMenu(menu)
        return menu



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import subprocess
import sys
import urllib.parse
import webbrowser
from typing import List

logger = logging.getLogger(__name__)


def compose_email(
    email: str, subject: str, body: str, attachment_filenames: List[str] | None = None, run_in_background=True
) -> None:
    # Encode the subject and body to ensure spaces and special characters are handled correctly
    subject_encoded = urllib.parse.quote(subject)
    body_encoded = urllib.parse.quote(body)
    mailto_link = f"mailto:{email}?subject={subject_encoded}&body={body_encoded}"

    # Function to attempt opening the mailto link with the OS's default email client
    def try_native_open(mailto_link) -> None:
        if sys.platform.startswith("linux"):
            # Linux: Use xdg-open to handle the mailto link
            subprocess.run(["xdg-open", mailto_link], check=True)
        elif sys.platform.startswith("darwin"):
            # macOS: Use open command to handle the mailto link
            subprocess.run(["open", mailto_link], check=True)
        elif sys.platform.startswith("win32"):
            # Windows: Use start command to handle the mailto link
            subprocess.run(["cmd", "/c", "start", "", mailto_link], check=True, shell=True)

    try:
        # Attempt to use the native OS command to open the email client
        try_native_open(mailto_link)
    except Exception as e:
        logger.debug(f"Failed to open the default email client using the OS native command: {e}")
        logger.debug("Attempting to open using webbrowser module...")
        # If the native OS command fails, fall back to using the webbrowser module
        webbrowser.open(mailto_link, new=0, autoraise=True)


if __name__ == "__main__":
    # Example usage
    compose_email(
        email="recipient@example.com", subject="Test Subject", body="This is a test body with spaces."
    )



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import datetime
import logging
import os
import shutil
from datetime import timedelta
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import bdkpython as bdk
from bitcoin_qr_tools.data import Data
from PyQt6.QtCore import Qt, QTimer, pyqtSignal
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (
    QApplication,
    QFileDialog,
    QHBoxLayout,
    QInputDialog,
    QMessageBox,
    QSplitter,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.fx import FX
from bitcoin_safe.gui.qt.extended_tabwidget import ExtendedTabWidget
from bitcoin_safe.gui.qt.label_syncer import LabelSyncer
from bitcoin_safe.gui.qt.my_treeview import SearchableTab, TreeViewWithToolbar
from bitcoin_safe.gui.qt.qt_wallet_base import QtWalletBase, SyncStatus
from bitcoin_safe.gui.qt.sync_tab import SyncTab
from bitcoin_safe.pythonbdk_types import Balance
from bitcoin_safe.threading_manager import TaskThread, ThreadingManager
from bitcoin_safe.util import Satoshis

from ...config import UserConfig
from ...execute_config import ENABLE_THREADING
from ...mempool import MempoolData
from ...signals import Signals, UpdateFilter, UpdateFilterReason, WalletSignals
from ...tx import TxBuilderInfos, TxUiInfos, short_tx_id
from ...wallet import (
    DeltaCacheListTransactions,
    DescriptorExportTools,
    ProtoWallet,
    Wallet,
    filename_clean,
    get_wallets,
)
from .address_list import AddressList, AddressListWithToolbar
from .bitcoin_quick_receive import BitcoinQuickReceive
from .category_list import CategoryEditor
from .descriptor_ui import DescriptorUI
from .dialogs import PasswordCreation, PasswordQuestion, question_dialog
from .hist_list import HistList, HistListWithToolbar
from .plot import WalletBalanceChart
from .taglist import AddressDragInfo
from .ui_tx import UITx_Creator
from .util import (
    Message,
    MessageType,
    caught_exception_message,
    custom_exception_handler,
    read_QIcon,
    save_file_dialog,
)
from .utxo_list import UTXOList, UtxoListWithToolbar

logger = logging.getLogger(__name__)


class QTProtoWallet(QtWalletBase):
    signal_create_wallet = pyqtSignal()
    signal_close_wallet = pyqtSignal()

    def __init__(
        self,
        protowallet: ProtoWallet,
        config: UserConfig,
        signals: Signals,
        get_lang_code: Callable[[], str],
        threading_parent: ThreadingManager | None = None,
    ) -> None:
        super().__init__(
            config=config, signals=signals, threading_parent=threading_parent, get_lang_code=get_lang_code
        )

        (
            self.wallet_descriptor_tab,
            self.wallet_descriptor_ui,
        ) = self.create_and_add_settings_tab(protowallet)

        self.tabs.setVisible(False)

    @property
    def protowallet(self) -> ProtoWallet:
        return self.wallet_descriptor_ui.protowallet

    @protowallet.setter
    def protowallet(self, protowallet) -> None:
        self.wallet_descriptor_ui.set_protowallet(protowallet)

    def get_mn_tuple(self) -> Tuple[int, int]:
        return self.protowallet.threshold, len(self.protowallet.keystores)

    def get_keystore_labels(self) -> List[str]:
        return [self.protowallet.signer_name(i) for i in range(len(self.protowallet.keystores))]

    def create_and_add_settings_tab(self, protowallet: ProtoWallet) -> Tuple[QWidget, DescriptorUI]:
        "Create a wallet settings tab, such that one can create a wallet (e.g. with xpub)"
        wallet_descriptor_ui = DescriptorUI(
            protowallet=protowallet,
            signals_min=self.signals,
            threading_parent=self,
            get_lang_code=self.get_lang_code,
        )
        self.tabs.add_tab(
            tab=wallet_descriptor_ui.tab,
            icon=read_QIcon("preferences.svg"),
            description=self.tr("Setup wallet"),
            data=wallet_descriptor_ui,
        )

        wallet_descriptor_ui.signal_qtwallet_apply_setting_changes.connect(self.on_apply_setting_changes)
        wallet_descriptor_ui.signal_qtwallet_cancel_wallet_creation.connect(self.signal_close_wallet.emit)
        return wallet_descriptor_ui.tab, wallet_descriptor_ui

    def on_apply_setting_changes(self) -> None:
        try:
            self.wallet_descriptor_ui.set_protowallet_from_ui()
        except Exception as e:
            Message(str(e), type=MessageType.Error)
            return

        self.signal_create_wallet.emit()

    def get_editable_protowallet(self) -> ProtoWallet:
        return self.protowallet


class QTWallet(QtWalletBase):
    signal_settext_balance_label = pyqtSignal(str)
    signal_on_change_sync_status = pyqtSignal(SyncStatus)  # SyncStatus

    def __init__(
        self,
        wallet: Wallet,
        config: UserConfig,
        signals: Signals,
        mempool_data: MempoolData,
        fx: FX,
        get_lang_code: Callable[[], str],
        set_tab_widget_icon: Optional[Callable[[QWidget, QIcon], None]] = None,
        password: str | None = None,
        file_path: str | None = None,
        threading_parent: ThreadingManager | None = None,
    ) -> None:
        super().__init__(
            signals=signals, config=config, threading_parent=threading_parent, get_lang_code=get_lang_code
        )
        self.mempool_data = mempool_data
        self.wallet = self.set_wallet(wallet)
        self.password = password
        self.set_tab_widget_icon = set_tab_widget_icon
        self.fx = fx
        self._file_path = file_path
        self.sync_status: SyncStatus = SyncStatus.unknown
        self.timer_sync_retry = QTimer()
        self.timer_sync_regularly = QTimer()

        self._last_syncing_start = datetime.datetime.now()
        self._syncing_delay = timedelta(seconds=0)
        self._last_sync_chain_height = 0

        ########### create tabs
        (
            self.history_tab,
            self.history_list,
            self.balance_plot,
            self.history_tab_with_toolbar,
        ) = self._create_hist_tab(self.tabs)

        (
            self.addresses_tab,
            self.address_list,
            self.address_list_tags,
        ) = self._create_addresses_tab(self.tabs)

        self.send_tab, self.uitx_creator = self._create_send_tab(self.tabs)

        (
            self.wallet_descriptor_tab,
            self.wallet_descriptor_ui,
        ) = self.create_and_add_settings_tab()

        self.sync_tab, self.sync_tab_widget, self.label_syncer = self.create_and_add_sync_tab()

        self.create_status_bar(self.tab, self.outer_layout)

        self.update_status_visualization(self.sync_status)
        self.tabs.setCurrentIndex(0)

        self.address_list.signal_tag_dropped.connect(self.set_category)
        self.address_list_tags.list_widget.signal_addresses_dropped.connect(self.set_category)
        self.address_list_tags.delete_button.signal_addresses_dropped.connect(self.set_category)
        self.address_list_tags.list_widget.signal_tag_deleted.connect(self.delete_category)
        self.address_list_tags.list_widget.signal_tag_renamed.connect(
            lambda old, new: self.rename_category(old, new)
        )

        self.updateUi()
        self.quick_receive.update_content(UpdateFilter(refresh_all=True))

        #### connect signals
        self.signal_on_change_sync_status.connect(self.update_status_visualization)
        self.signals.language_switch.connect(self.updateUi)
        self.wallet_signals.updated.connect(self.signals.any_wallet_updated.emit)
        self.wallet_signals.export_labels.connect(self.export_labels)
        self.wallet_signals.export_bip329_labels.connect(self.export_bip329_labels)
        self.wallet_signals.import_labels.connect(self.import_labels)
        self.wallet_signals.import_bip329_labels.connect(self.import_bip329_labels)
        self.wallet_signals.import_electrum_wallet_labels.connect(self.import_electrum_wallet_labels)
        self.signal_on_change_sync_status.connect(self.update_display_balance)

        self._start_sync_retry_timer()
        self._start_sync_regularly_timer()

    @property
    def wallet_signals(self) -> WalletSignals:
        return self.signals.wallet_signals[self.wallet.id]

    def updateUi(self) -> None:
        self.tabs.setTabText(self.tabs.indexOf(self.send_tab), self.tr("Send"))
        self.tabs.setTabText(self.tabs.indexOf(self.wallet_descriptor_tab), self.tr("Descriptor"))
        self.tabs.setTabText(self.tabs.indexOf(self.sync_tab_widget), self.tr("SyncTalk"))
        self.tabs.setTabText(self.tabs.indexOf(self.history_tab), self.tr("History"))
        self.tabs.setTabText(self.tabs.indexOf(self.addresses_tab), self.tr("Receive"))

    def set_display_balance(self, value: Balance):
        self.wallet.data_dump["display_balance"] = value

    def get_display_balance(self) -> Balance:
        if self.sync_status == SyncStatus.synced:
            return self.wallet.get_balance()
        else:
            return self.wallet.data_dump.get("display_balance", Balance())

    def update_display_balance(self):
        if self.sync_status == SyncStatus.synced:
            self.set_display_balance(self.wallet.get_balance())

    def stop_sync_timer(self) -> None:
        self.timer_sync_retry.stop()
        self.timer_sync_regularly.stop()

    def close(self) -> None:
        self.disconnect_signals()
        self.label_syncer.send_all_labels_to_myself()
        self.sync_tab.unsubscribe_all()
        self.sync_tab.nostr_sync.stop()
        self.stop_sync_timer()
        self.end_threading_manager()

    def _start_sync_regularly_timer(self, delay_retry_sync=60) -> None:
        if self.timer_sync_regularly.isActive():
            return
        self.timer_sync_regularly.setInterval(delay_retry_sync * 1000)

        def sync() -> None:
            if self.sync_status not in [SyncStatus.synced]:
                return

            logger.info(f"Regular update: Sync wallet {self.wallet.id} again")
            self.sync()

        self.timer_sync_regularly.timeout.connect(sync)
        self.timer_sync_regularly.start()

    def _start_sync_retry_timer(self, delay_retry_sync=30) -> None:
        if self.timer_sync_retry.isActive():
            return
        self.timer_sync_retry.setInterval(delay_retry_sync * 1000)

        def sync_if_needed() -> None:
            if self.sync_status in [SyncStatus.syncing, SyncStatus.synced]:
                return

            logger.info(f"Retry timer: Try syncing wallet {self.wallet.id}")
            self.sync()

        self.timer_sync_retry.timeout.connect(sync_if_needed)
        self.timer_sync_retry.start()

    def get_mn_tuple(self) -> Tuple[int, int]:
        return self.wallet.get_mn_tuple()

    def get_keystore_labels(self) -> List[str]:
        return [keystore.label for keystore in self.wallet.keystores]

    @property
    def file_path(self) -> str:
        return self._file_path if self._file_path else filename_clean(self.wallet.id)

    @file_path.setter
    def file_path(self, value: Optional[str]) -> None:
        self._file_path = value

    def create_and_add_settings_tab(self) -> Tuple[QWidget, DescriptorUI]:
        "Create a wallet settings tab, such that one can create a wallet (e.g. with xpub)"
        wallet_descriptor_ui = DescriptorUI(
            protowallet=self.wallet.as_protowallet(),
            signals_min=self.signals,
            get_wallet=lambda: self.wallet,
            get_lang_code=self.get_lang_code,
        )
        self.tabs.add_tab(
            tab=wallet_descriptor_ui.tab,
            icon=read_QIcon("preferences.svg"),
            description="",
            data=wallet_descriptor_ui,
        )

        wallet_descriptor_ui.signal_qtwallet_apply_setting_changes.connect(
            self.on_qtwallet_apply_setting_changes
        )
        wallet_descriptor_ui.signal_qtwallet_cancel_setting_changes.connect(self.cancel_setting_changes)
        return wallet_descriptor_ui.tab, wallet_descriptor_ui

    def on_qtwallet_apply_setting_changes(self):
        # save old status, such that the backup has all old data (inlcuding the "SyncTab" in the data_dump)
        self.save()

        self.wallet_descriptor_ui.set_protowallet_from_ui()
        new_wallet = Wallet.from_protowallet(
            protowallet=self.wallet_descriptor_ui.protowallet,
            config=self.config,
            data_dump=None,  # This forces a recrtion of the sync tab, including the updated group-discovery key from the (updated) descriptor  #,self.wallet.data_dump,
            labels=self.wallet.labels,
            default_category=self.wallet.labels.default_category,
        )
        # compare if something change
        if self.wallet.is_essentially_equal(new_wallet):
            Message(self.tr("No changes to apply."))
            return

        # do backup
        filename = self.save_backup()
        if filename:
            Message(self.tr("Backup saved to {filename}").format(filename=filename))
        else:
            Message(self.tr("Backup failed. Aborting Changes."))
            return

        if not question_dialog(
            self.tr("Proceeding will potentially change all wallet addresses. Do you want to proceed?")
        ):
            return

        self.signals.close_qt_wallet.emit(self.wallet.id)
        self.signals.create_qt_wallet_from_wallet.emit(new_wallet, self._file_path, self.password)

    def get_sync_tab_data(self) -> Optional[Dict]:
        return self.wallet.data_dump.get("SyncTab")

    def set_sync_tab_data(self, data: Dict):
        self.wallet.data_dump["SyncTab"] = data

    def create_and_add_sync_tab(self) -> Tuple[SyncTab, QWidget, LabelSyncer]:
        "Create a wallet settings tab, such that one can create a wallet (e.g. with xpub)"
        sync_tab_data = self.get_sync_tab_data()
        sync_tab = (
            SyncTab.from_dump(sync_tab_data, network=self.config.network, signals=self.signals)
            if sync_tab_data
            else SyncTab.from_descriptor_new_device_keys(
                self.wallet.multipath_descriptor,
                network=self.config.network,
                signals=self.signals,
                parent=self,
            )
        )
        self.set_sync_tab_data(sync_tab.dump())

        self.tabs.add_tab(
            tab=sync_tab.main_widget, icon=read_QIcon("cloud-sync.svg"), description="", data=sync_tab
        )

        label_syncer = LabelSyncer(self.wallet.labels, sync_tab, self.wallet_signals)
        sync_tab.finish_init_after_signal_connection()
        return sync_tab, sync_tab.main_widget, label_syncer

    def __repr__(self) -> str:
        return f"QTWallet({self.__dict__})"

    def save_backup(self) -> str:
        """_summary_

        Returns:
            str: filename
        """
        filename = os.path.join(
            self.config.wallet_dir, "backups", filename_clean(f"{self.wallet.id}-backup-{0}")
        )
        max_number_backups = 100000
        for i in range(max_number_backups):
            filename = os.path.join(
                self.config.wallet_dir, "backups", filename_clean(f"{self.wallet.id}-backup-{i}")
            )
            if not os.path.exists(filename):
                break

        # save the tutorial step into the wallet
        if self.wallet_steps:
            self.wallet.tutorial_index = (
                self.wallet_steps.current_index() if not self.wallet_steps.isHidden() else None
            )

        self.wallet.save(
            filename,
            password=self.password,
        )
        return filename

    def move_wallet_file(self, new_file_path) -> Optional[str]:
        if os.path.exists(new_file_path):
            Message(
                self.tr("Cannot move the wallet file, because {file_path} exists").format(
                    file_path=new_file_path
                )
            )
            return None
        shutil.move(self.file_path, new_file_path)
        self.remove_lockfile(Path(self.file_path))
        old_file_path = self.file_path
        self.file_path = new_file_path
        self.get_wallet_lockfile(Path(self.file_path))
        logger.info(f"Saved {old_file_path} under new name {self.file_path}")
        return new_file_path

    @classmethod
    def get_wallet_lockfile_path(cls, wallet_file_path: Path) -> Path:
        return wallet_file_path.with_suffix(".lock")

    @classmethod
    def get_wallet_lockfile(cls, wallet_file_path: Path) -> Optional[Path]:
        lockfile_path = cls.get_wallet_lockfile_path(wallet_file_path)
        if os.path.exists(lockfile_path):
            return None
        with open(lockfile_path, "w") as lockfile:
            lockfile.write(str(os.getpid()))
            return lockfile_path

    @classmethod
    def remove_lockfile(cls, wallet_file_path: Path) -> None:
        lock_file_path = cls.get_wallet_lockfile_path(wallet_file_path)
        if not lock_file_path:
            return
        if lock_file_path.exists():
            os.remove(lock_file_path)
            logger.info(f"Lock file {lock_file_path} removed.")

    def save(self) -> Optional[str]:
        if not self._file_path:
            if not os.path.exists(self.config.wallet_dir):
                os.makedirs(self.config.wallet_dir, exist_ok=True)

            # not saving a wallet is dangerous. Therefore I ensure the user has ample
            # opportunity to set a filename
            while not self._file_path:
                self._file_path, _ = QFileDialog.getSaveFileName(
                    self.tab,
                    self.tr("Save wallet"),
                    f"{os.path.join(self.config.wallet_dir, filename_clean(self.wallet.id))}",
                    self.tr("All Files (*);;Wallet Files (*.wallet)"),
                )
                if not self._file_path and question_dialog(
                    text=self.tr("Are you SURE you don't want save the wallet {id}?").format(
                        id=self.wallet.id
                    ),
                    title=self.tr("Delete wallet"),
                ):
                    logger.info("No file selected")
                    return None

        # if it is the first time saving, then the user can set a password
        if not os.path.isfile(self.file_path):
            self.password = PasswordCreation().get_password()

        # save the tutorial step into the wallet
        if self.wallet_steps:
            self.wallet.tutorial_index = (
                self.wallet_steps.current_index() if not self.wallet_steps.isHidden() else None
            )

        if self.sync_tab:
            self.set_sync_tab_data(self.sync_tab.dump())

        self.wallet.save(
            self.file_path,
            password=self.password,
        )
        logger.info(f"wallet {self.wallet.id} saved")
        return self.file_path

    def change_password(self) -> Optional[str]:
        if self.password:
            ui_password_question = PasswordQuestion(label_text="Your current password:")
            password = ui_password_question.ask_for_password()
            password = password if password else None
            if password != self.password:
                Message(self.tr("Password incorrect"), type=MessageType.Warning)
                return None

        self.password = PasswordCreation(
            window_title=self.tr("Change password"), label_text=self.tr("New password:")
        ).get_password()
        self.save()
        Message(self.tr("Wallet saved"))
        return self.password

    def cancel_setting_changes(self) -> None:
        self.wallet_descriptor_ui.protowallet = self.wallet.as_protowallet()
        self.wallet_descriptor_ui.set_all_ui_from_protowallet()

    def get_delta_txs(self, access_marker="notifications") -> DeltaCacheListTransactions:
        delta_txs = self.wallet.bdkwallet.list_delta_transactions(access_marker=access_marker)
        return delta_txs

    def format_txs(self, txs: List[bdk.TransactionDetails]) -> str:
        return "\n".join(
            [
                self.tr("  {amount} in {shortid}").format(
                    amount=Satoshis(tx.received - tx.sent, self.config.network).str_as_change(unit=True),
                    shortid=short_tx_id(tx.txid),
                )
                for tx in txs
            ]
        )

    def hanlde_removed_txs(self, removed_txs: List[bdk.TransactionDetails]) -> None:
        if not removed_txs:
            return

        # if transactions were removed (reorg or other), then recalculate everything
        message_content = self.tr(
            "The transactions \n{txs}\n in wallet '{wallet}' were removed from the history!!!"
        ).format(txs=self.format_txs(removed_txs), wallet=self.wallet.id)
        Message(
            message_content,
            no_show=True,
        ).emit_with(self.signals.notification)
        if question_dialog(
            message_content + "\n" + self.tr("Do you want to save a copy of these transactions?")
        ):
            folder_path = QFileDialog.getExistingDirectory(
                self.tab, "Select Folder to save the removed transactions"
            )

            if folder_path:
                for tx in removed_txs:
                    data = Data.from_tx(tx.transaction)
                    filename = Path(folder_path) / f"{short_tx_id( tx.txid)}.tx"

                    # create a file descriptor
                    fd = os.open(filename, os.O_CREAT | os.O_WRONLY)
                    data.write_to_filedescriptor(fd)
                    logger.info(f"Exported {tx.txid} to {filename}")

        # all the lists must be updated
        self.refresh_caches_and_ui_lists(force_ui_refresh=True)

    def handle_appended_txs(self, appended_txs: List[bdk.TransactionDetails]) -> None:
        if not appended_txs:
            return

        if len(appended_txs) == 1:
            Message(
                self.tr("New transaction in wallet '{wallet}':\n{txs}").format(
                    txs=self.format_txs(appended_txs), wallet=self.wallet.id
                ),
                no_show=True,
            ).emit_with(self.signals.notification)
        else:
            Message(
                self.tr("{number} new transactions in wallet '{wallet}':\n{txs}").format(
                    number=len(appended_txs),
                    txs=self.format_txs(appended_txs),
                    wallet=self.wallet.id,
                ),
                no_show=True,
            ).emit_with(self.signals.notification)

    def handle_delta_txs(self, delta_txs: DeltaCacheListTransactions) -> None:
        self.hanlde_removed_txs(delta_txs.removed)
        self.handle_appended_txs(delta_txs.appended)

    def refresh_caches_and_ui_lists(
        self,
        enable_threading=ENABLE_THREADING,
        force_ui_refresh=True,
        chain_height_advanced=False,
        clear_cache=True,
    ) -> None:
        # before the wallet UI updates, we have to refresh the wallet caches to make the UI update faster
        logger.debug("refresh_caches_and_ui_lists")
        if clear_cache:
            self.wallet.clear_cache()

        def do() -> Any:
            self.wallet.fill_commonly_used_caches()

        def on_done(result) -> None:
            delta_txs = self.get_delta_txs()
            change_dict = delta_txs.was_changed()
            logger.debug(f"change_dict={change_dict}")
            if force_ui_refresh or change_dict:
                self.handle_delta_txs(delta_txs)

                # now do the UI
                logger.debug("start refresh ui")

                self.wallet_signals.updated.emit(
                    UpdateFilter(
                        refresh_all=True,
                        reason=(
                            UpdateFilterReason.TransactionChange
                            if change_dict
                            else UpdateFilterReason.ForceRefresh
                        ),
                    )
                )
            elif chain_height_advanced:
                self.wallet_signals.updated.emit(UpdateFilter(reason=UpdateFilterReason.ChainHeightAdvanced))

        def on_success(result) -> None:
            # now do the UI
            logger.debug("on_success refresh ui")

        def on_error(packed_error_info) -> None:
            custom_exception_handler(*packed_error_info)

        self.append_thread(
            TaskThread(enable_threading=enable_threading).add_and_start(do, on_success, on_done, on_error)
        )

    def _create_send_tab(self, tabs: QTabWidget) -> Tuple[SearchableTab, UITx_Creator]:
        utxo_list = UTXOList(
            self.config,
            self.signals,
            get_outpoints=lambda: [],  # this is filled in uitx_creator
            hidden_columns=[
                UTXOList.Columns.OUTPOINT,
                UTXOList.Columns.PARENTS,
                UTXOList.Columns.WALLET_ID,
            ],
            sort_column=UTXOList.Columns.CATEGORY,
            sort_order=Qt.SortOrder.AscendingOrder,
        )

        toolbar_list = UtxoListWithToolbar(utxo_list, self.config, parent=tabs)

        uitx_creator = UITx_Creator(
            wallet=self.wallet,
            mempool_data=self.mempool_data,
            fx=self.fx,
            categories=self.wallet.labels.categories,
            widget_utxo_with_toolbar=toolbar_list,
            utxo_list=utxo_list,
            config=self.config,
            signals=self.signals,
            parent=self.tab,
        )
        self.tabs.add_tab(tab=uitx_creator, icon=read_QIcon("send.svg"), description="", data=uitx_creator)

        uitx_creator.signal_create_tx.connect(self.create_psbt)

        return uitx_creator, uitx_creator

    def create_psbt(self, txinfos: TxUiInfos) -> None:

        def do() -> Union[TxBuilderInfos, Exception]:
            try:
                return self.wallet.create_psbt(txinfos)
            except Exception as e:
                return e

        def on_done(builder_infos: Union[TxBuilderInfos, Exception]) -> None:
            if not builder_infos:
                self.wallet_signals.finished_psbt_creation.emit()
                return
            if isinstance(builder_infos, Exception):
                caught_exception_message(builder_infos)
                self.wallet_signals.finished_psbt_creation.emit()
                return
            if not isinstance(builder_infos, TxBuilderInfos):
                self.wallet_signals.finished_psbt_creation.emit()  # type: ignore
                Message("Could not create PSBT", type=MessageType.Error)
                return

            try:
                # set labels in other wallets  (recipients can be another open wallet)
                for wallet in get_wallets(self.signals):
                    wallet._set_labels_for_change_outputs(builder_infos)

                update_filter = UpdateFilter(
                    addresses=set(
                        [
                            bdk.Address.from_script(output.script_pubkey, self.wallet.network).as_string()
                            for output in builder_infos.builder_result.psbt.extract_tx().output()
                        ]
                    ),
                    reason=UpdateFilterReason.CreatePSBT,
                )
                self.wallet_signals.updated.emit(update_filter)
                self.signals.open_tx_like.emit(builder_infos)

                self.uitx_creator.clear_ui()
            except Exception as e:
                caught_exception_message(e)
            finally:
                self.wallet_signals.finished_psbt_creation.emit()

        def on_success(builder_infos: Union[TxBuilderInfos, Exception]) -> None:
            pass

        def on_error(packed_error_info) -> None:
            self.wallet_signals.finished_psbt_creation.emit()

        self.append_thread(TaskThread().add_and_start(do, on_success, on_done, on_error))

    def set_wallet(self, wallet: Wallet) -> Wallet:
        self.wallet = wallet

        self.connect_signal(self.wallet_signals.updated, self.wallet.on_addresses_updated)

        self.connect_signal(self.signals.get_wallets, lambda: self.wallet, slot_name=self.wallet.id)
        self.connect_signal(self.signals.get_qt_wallets, lambda: self, slot_name=self.wallet.id)
        self.connect_signal(
            self.wallet_signals.get_display_balance, self.get_display_balance, slot_name=self.wallet.id
        )
        return wallet

    def rename_category(self, old_category: str, new_category: str) -> None:
        affected_keys = self.wallet.labels.rename_category(old_category, new_category)
        self.wallet_signals.updated.emit(
            UpdateFilter(
                addresses=affected_keys,
                categories=([old_category]),
                txids=affected_keys,
                reason=UpdateFilterReason.CategoryRenamed,
            )
        )

    def delete_category(self, category: str) -> None:
        # Show dialog and get input
        text, ok = QInputDialog.getText(
            self.tab,
            self.tr("Rename or merge categories"),
            self.tr("Choose a new name, or an existing name for merging:"),
            text=category,
        )
        if ok and text:
            new_category = text
        else:
            # just rename it with the same name
            new_category = category

        affected_keys = self.wallet.labels.rename_category(category, new_category)
        self.wallet_signals.updated.emit(
            UpdateFilter(
                addresses=affected_keys,
                categories=([category]),
                txids=affected_keys,
                reason=UpdateFilterReason.CategoryRenamed,
            )
        )

    def set_category(self, address_drag_info: AddressDragInfo) -> None:
        for address in address_drag_info.addresses:
            for category in address_drag_info.tags:
                self.wallet.labels.set_addr_category(address, category, timestamp="now")

        txids: Set[str] = set()
        for address in address_drag_info.addresses:
            txids = txids.union(self.wallet.get_involved_txids(address))

        self.wallet_signals.updated.emit(
            UpdateFilter(
                addresses=address_drag_info.addresses,
                categories=address_drag_info.tags,
                txids=txids,
                reason=UpdateFilterReason.UserInput,
            )
        )

    def create_status_bar(self, tab: QWidget, outer_layout) -> None:
        pass

    def update_status_visualization(self, sync_status: SyncStatus) -> None:
        if not self.wallet:
            return

        icon = None
        if sync_status == SyncStatus.syncing:
            icon = read_QIcon("status_waiting.svg")
            self.history_tab_with_toolbar.sync_button.set_icon_is_syncing()
        elif self.wallet.get_height() and sync_status in [SyncStatus.synced]:
            icon = read_QIcon("status_connected.svg")
            self.history_tab_with_toolbar.sync_button.set_icon_allow_refresh()
        else:
            icon = read_QIcon("status_disconnected.svg")
            self.history_tab_with_toolbar.sync_button.set_icon_allow_refresh()

        if self.set_tab_widget_icon:
            self.set_tab_widget_icon(self.tab, icon)

    def create_list_tab(
        self,
        toolbar: TreeViewWithToolbar,
        tabs: QTabWidget,
        horizontal_widgets_left: Optional[List[QWidget]] = None,
        horizontal_widgets_right: Optional[List[QWidget]] = None,
    ) -> SearchableTab:
        # create a horizontal widget and layout
        h = SearchableTab(parent=tabs)
        h.searchable_list = toolbar.searchable_list
        hbox = QHBoxLayout(h)
        h.setLayout(hbox)

        if horizontal_widgets_left:
            for widget in horizontal_widgets_left:
                hbox.addWidget(widget)

        hbox.addWidget(toolbar)

        if horizontal_widgets_right:
            for widget in horizontal_widgets_right:
                hbox.addWidget(widget)
        return h

    def _create_hist_tab(
        self, tabs: ExtendedTabWidget
    ) -> Tuple[SearchableTab, HistList, WalletBalanceChart, HistListWithToolbar]:
        tab = SearchableTab(parent=tabs)
        tab_layout = QHBoxLayout(tab)

        splitter1 = QSplitter()  # horizontal splitter by default
        tab_layout.addWidget(splitter1)

        l = HistList(
            fx=self.fx,
            config=self.config,
            signals=self.signals,
            wallet_id=self.wallet.id,
            hidden_columns=[
                HistList.Columns.WALLET_ID,
                HistList.Columns.BALANCE,
                HistList.Columns.TXID,
            ],
            mempool_data=self.mempool_data,
        )
        toolbar = HistListWithToolbar(l, self.config, parent=tabs)

        list_widget = self.create_list_tab(toolbar, tabs)
        splitter1.addWidget(list_widget)
        tab.searchable_list = l

        right_widget = QWidget()
        right_widget_layout = QVBoxLayout(right_widget)
        right_widget_layout.setContentsMargins(0, 0, 0, 0)

        self.quick_receive: BitcoinQuickReceive = BitcoinQuickReceive(self.wallet_signals, self.wallet)
        right_widget_layout.addWidget(self.quick_receive)

        plot = WalletBalanceChart(self.wallet, wallet_signals=self.wallet_signals)
        right_widget_layout.addWidget(plot)

        splitter1.addWidget(right_widget)

        tabs.add_tab(
            tab=tab, icon=read_QIcon("history.svg"), description="", position=2, data=[toolbar, plot]
        )

        splitter1.setSizes([1, 1])
        return tab, l, plot, toolbar

    def _subtexts_for_categories(self) -> List[str]:
        return [self.tr("Click for new address")] * len(self.wallet.labels.categories)

        # d = {}
        # for address in self.wallet.get_addresses():
        #     category = self.wallet.labels.get_category(address)
        #     if category not in d:
        #         d[category] = []

        #     d[category].append(address)

        # return [f"{len(d.get(category, []))} Addresses" for category in self.wallet.labels.categories]

    def _create_addresses_tab(
        self, tabs: ExtendedTabWidget
    ) -> Tuple[SearchableTab, AddressList, CategoryEditor]:
        l = AddressList(
            fx=self.fx,
            config=self.config,
            wallet=self.wallet,
            wallet_signals=self.wallet_signals,
            signals=self.signals,
        )
        toolbar = AddressListWithToolbar(l, self.config, parent=tabs, signals=self.signals)

        tags = CategoryEditor(
            lambda: self.wallet.labels.categories,
            self.wallet_signals,
            get_sub_texts=self._subtexts_for_categories,
        )
        tags.signal_category_added.connect(self.on_add_category)

        def create_new_address(category) -> None:
            self.address_list.get_address(force_new=True, category=category)

        tags.list_widget.signal_tag_clicked.connect(create_new_address)

        tags.setMaximumWidth(150)
        tab = self.create_list_tab(toolbar, tabs, horizontal_widgets_left=[tags])

        tabs.add_tab(tab=tab, icon=read_QIcon("receive.svg"), description="", position=1, data=toolbar)
        return tab, l, tags

    def on_add_category(self, category: str) -> None:
        self.wallet.labels.add_category(category)

    def set_sync_status(self, new: SyncStatus) -> None:
        self.sync_status = new
        logger.info(f"{self.wallet.id} set_sync_status {new}")
        self.signal_on_change_sync_status.emit(new)
        QApplication.processEvents()

    def sync(self) -> None:
        if self.sync_status == SyncStatus.syncing:
            logger.info(f"Syncing already in progress")
            return

        def progress_function_threadsafe(progress: float, message: str) -> None:
            self.signal_settext_balance_label.emit(f"Syncing wallet: {round(progress)}%  {message}")

        def do() -> Any:
            self.wallet.sync(progress_function_threadsafe=progress_function_threadsafe)

        def on_done(result) -> None:
            self._syncing_delay = datetime.datetime.now() - self._last_syncing_start
            interval_timer_sync_regularly = max(int(self._syncing_delay.total_seconds() * 200), 30)  # in sec
            self.timer_sync_regularly.setInterval(interval_timer_sync_regularly * 1000)
            logger.info(
                f"Syncing took {self._syncing_delay} --> set the interval_timer_sync_regularly to {interval_timer_sync_regularly}s"
            )

        def on_error(packed_error_info) -> None:
            self.set_sync_status(SyncStatus.error)
            logger.info(
                f"Could not sync. SynStatus set to {SyncStatus.error.name} for wallet {self.wallet.id}"
            )
            logger.error(str(packed_error_info))
            # custom_exception_handler(*packed_error_info)

        def on_success(result) -> None:
            self.set_sync_status(SyncStatus.synced)
            logger.info(f"{self.wallet.id} success syncing wallet {self.wallet.id}")

            logger.info("start updating lists")
            # self.wallet.clear_cache()
            self.refresh_caches_and_ui_lists(
                force_ui_refresh=False,
                chain_height_advanced=self.wallet.get_height() != self._last_sync_chain_height,
            )
            # self.update_tabs()
            logger.info("finished updating lists")
            self._last_sync_chain_height = self.wallet.get_height()

            self.fx.update_if_needed()
            self.signal_after_sync.emit(self.sync_status)

        logger.info(f"Refresh all caches before syncing.")
        # This takkles the following problem:
        # During the syncing process the cache and the bdk results
        # become inconsitent (since the bdk has newer info)
        # If now some results are caches, and some are not
        # then the results become inconsitent
        # Possible solutions:
        #   1. Disable cache while syncing and force access to bdk.
        #       This might work, but also probably has considerable freezing effects
        #       (since bdk is blocked due to syncing)
        #   2. Only use the cache during syncing.  This is the chosen solution here.
        #       by filling all the cache before the sync
        #       I do this in the main thread so all caches
        #       are filled before the syncing process
        self.refresh_caches_and_ui_lists(enable_threading=False, force_ui_refresh=False, clear_cache=False)

        logger.info(f"Start syncing wallet {self.wallet.id}")
        self.set_sync_status(SyncStatus.syncing)

        self._last_syncing_start = datetime.datetime.now()
        self.append_thread(TaskThread().add_and_start(do, on_success, on_done, on_error))

    def export_wallet_for_coldcard(self) -> Optional[str]:
        filename = save_file_dialog(
            name_filters=["Text (*.txt)", "All Files (*.*)"],
            default_suffix="txt",
            default_filename=filename_clean(self.wallet.id, file_extension=".txt")[:24],
        )
        if not filename:
            return None

        with open(filename, "w") as file:
            file.write(
                DescriptorExportTools.get_coldcard_str(self.wallet.id, self.wallet.multipath_descriptor)
            )
        return filename

    def get_editable_protowallet(self) -> ProtoWallet:
        return self.wallet.as_protowallet()

    def export_bip329_labels(self) -> None:
        s = self.wallet.labels.export_bip329_jsonlines()
        file_path, _ = QFileDialog.getSaveFileName(
            self.tab,
            self.tr("Export labels"),
            f"{self.wallet.id}_labels.jsonl",
            self.tr("All Files (*);;JSON Files (*.jsonl);;JSON Files (*.json)"),
        )
        if not file_path:
            logger.info("No file selected")
            return

        with open(file_path, "w") as file:
            file.write(s)

    def export_labels(self) -> None:
        s = self.wallet.labels.dumps_data_jsonlines()
        file_path, _ = QFileDialog.getSaveFileName(
            self.tab,
            self.tr("Export labels"),
            f"{self.wallet.id}_labels.jsonl",
            self.tr("All Files (*);;JSON Files (*.jsonl);;JSON Files (*.json)"),
        )
        if not file_path:
            logger.info("No file selected")
            return

        with open(file_path, "w") as file:
            file.write(s)

    def import_bip329_labels(self) -> None:
        file_path, _ = QFileDialog.getOpenFileName(
            self.tab,
            self.tr("Import labels"),
            "",
            self.tr("All Files (*);;JSONL Files (*.jsonl);;JSON Files (*.json)"),
        )
        if not file_path:
            logger.info("No file selected")
            return

        with open(file_path, "r") as file:
            lines = file.read()

        changed_data = self.wallet.labels.import_bip329_jsonlines(lines)
        self.wallet_signals.updated.emit(UpdateFilter(refresh_all=True, reason=UpdateFilterReason.UserImport))
        Message(
            self.tr("Successfully updated {number} Labels").format(number=len(changed_data)),
            type=MessageType.Info,
        )

    def import_labels(self) -> None:
        file_path, _ = QFileDialog.getOpenFileName(
            self.tab,
            self.tr("Import labels"),
            "",
            self.tr("All Files (*);;JSONL Files (*.jsonl);;JSON Files (*.json)"),
        )
        if not file_path:
            logger.info("No file selected")
            return

        with open(file_path, "r") as file:
            lines = file.read()

        force_overwrite = bool(
            question_dialog(
                "Do you want to overwrite all labels?",
                buttons=QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
            )
        )

        changed_data = self.wallet.labels.import_dumps_data(lines, force_overwrite=force_overwrite)
        self.wallet_signals.updated.emit(UpdateFilter(refresh_all=True, reason=UpdateFilterReason.UserImport))
        Message(
            self.tr("Successfully updated {number} Labels").format(number=len(changed_data)),
            type=MessageType.Info,
        )

    def import_electrum_wallet_labels(self) -> None:

        file_path, _ = QFileDialog.getOpenFileName(
            self.tab,
            self.tr("Import Electrum Wallet labels"),
            "",
            self.tr("All Files (*);;JSON Files (*.json)"),
        )
        if not file_path:
            logger.info("No file selected")
            return

        with open(file_path, "r") as file:
            lines = file.read()

        changed_data = self.wallet.labels.import_electrum_wallet_json(lines, network=self.config.network)
        self.wallet_signals.updated.emit(UpdateFilter(refresh_all=True, reason=UpdateFilterReason.UserImport))
        Message(
            self.tr("Successfully updated {number} Labels").format(number=len(changed_data)),
            type=MessageType.Info,
        )



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import Any, Dict, List, Literal, Optional, Set

logger = logging.getLogger(__name__)

import copy

import bdkpython as bdk
from bitcoin_usb.address_types import (
    AddressType,
    AddressTypes,
    ConstDerivationPaths,
    SimplePubKeyProvider,
)
from packaging import version

from .descriptors import MultipathDescriptor
from .storage import BaseSaveableClass, SaveAllClass, filtered_for_init


class KeyStoreImporterType(SaveAllClass):
    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        icon_filename: str,
        networks: List[bdk.Network] | Literal["all"] = "all",
    ) -> None:
        self.id = id
        self.name = name
        self.description = description
        self.icon_filename = icon_filename
        self.networks = (
            [
                bdk.Network.BITCOIN,
                bdk.Network.REGTEST,
                bdk.Network.TESTNET,
                bdk.Network.SIGNET,
            ]
            if networks == "all"
            else networks
        )

    @classmethod
    def from_dump_migration(cls, dct: Dict[str, Any]) -> Dict[str, Any]:
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.0"):
            pass

        # now the version is newest, so it can be deleted from the dict
        if "VERSION" in dct:
            del dct["VERSION"]
        return dct


class KeyStoreImporterTypes:
    hwi = KeyStoreImporterType("hwi", "USB hardware signer", "Connect \nUSB \nhardware signer", "usb.svg")
    file = KeyStoreImporterType(
        "file",
        "SD card",
        "Import signer details\nvia SD card",
        "sd-card.svg",
    )
    clipboard = KeyStoreImporterType(
        "clipboard",
        "Clipboard",
        "Import signer details\nfrom text",
        "clip.svg",
    )
    qr = KeyStoreImporterType(
        "qr",
        "QR Code",
        "Import signer details\nvia QR code",
        "camera.svg",
    )
    watch_only = KeyStoreImporterType(
        "watch_only",
        "Watch-Only",
        "xPub / Public Key\nInformation",
        "key-hole-icon.svg",
    )
    seed = KeyStoreImporterType(
        "seed",
        "Seed",
        "Mnemonic Seed\n(Testnet only)",
        "logo.svg",
        networks=[bdk.Network.REGTEST, bdk.Network.TESTNET, bdk.Network.SIGNET],
    )  # add networks here to make the seed option visible

    @classmethod
    def list_types(cls, network: bdk.Network) -> List[KeyStoreImporterType]:
        return [v for v in [cls.hwi, cls.file, cls.qr, cls.watch_only, cls.seed] if network in v.networks]

    @classmethod
    def list_names(cls, network: bdk.Network) -> List[str]:
        return [v.name for v in cls.list_types(network)]


class KeyStore(SimplePubKeyProvider, BaseSaveableClass):
    VERSION = "0.0.2"
    known_classes = {
        **BaseSaveableClass.known_classes,
    }

    def __init__(
        self,
        xpub: str,
        fingerprint: str,
        key_origin: str,
        label: str,
        network: bdk.Network,
        mnemonic: Optional[str] = None,
        description: str = "",
        derivation_path: str = ConstDerivationPaths.receive,
    ) -> None:
        super().__init__(
            xpub=xpub,
            fingerprint=fingerprint,
            key_origin=key_origin,
            derivation_path=derivation_path,
        )

        self.network = network
        if not self.is_xpub_valid(xpub=xpub, network=self.network):
            raise ValueError(f"{xpub} is not a valid xpub")

        self.label = label
        self.mnemonic = mnemonic
        self.description = description

    def get_relevant_differences(self, other_keystore: "KeyStore") -> Set[str]:
        "Compares the relevant entries like keystores"
        differences = set()
        this = self.dump()
        other = other_keystore.dump()

        keys = [
            "xpub",
            "fingerprint",
            "key_origin",
            "network",
            "mnemonic",
            "derivation_path",
        ]
        for k in keys:
            if this[k] != other[k]:
                differences.add(k)

        return differences

    def is_equal(self, other: "KeyStore") -> bool:
        return self.__dict__ == other.__dict__

    @classmethod
    def is_seed_valid(cls, mnemonic: str) -> bool:
        try:
            bdk.Mnemonic.from_string(mnemonic)
            return True
        except:
            return False

    @classmethod
    def is_xpub_valid(cls, xpub: str, network: bdk.Network) -> bool:
        try:
            AddressTypes.p2pkh.bdk_descriptor(
                bdk.DescriptorPublicKey.from_string(xpub),
                "0" * 8,
                bdk.KeychainKind.EXTERNAL,
                network,
            )

            return True
        except:
            return False

    def clone(self, class_kwargs=None) -> "KeyStore":
        return KeyStore(**self.__dict__)

    def to_singlesig_multipath_descriptor(
        self, address_type: AddressType, network: bdk.Network
    ) -> MultipathDescriptor:
        "Uses the bdk descriptor templates to create the descriptor from xpub or seed"
        descriptors = [
            (
                address_type.bdk_descriptor(
                    bdk.DescriptorPublicKey.from_string(self.xpub),
                    self.fingerprint,
                    keychainkind,
                    network,
                )
                if not self.mnemonic
                else address_type.bdk_descriptor_secret(
                    bdk.DescriptorSecretKey(network, bdk.Mnemonic.from_str(self.mnemonic), ""),  # type: ignore
                    keychainkind,
                    network,
                )
            )
            for keychainkind in [
                bdk.KeychainKind.EXTERNAL,
                bdk.KeychainKind.INTERNAL,
            ]
        ]
        return MultipathDescriptor(*descriptors)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    def dump(self) -> Dict[str, Any]:
        d = super().dump()

        # you must copy it, so you not't change any calues
        full_dict = self.__dict__.copy()
        # the deepcopy must be done AFTER there is no bdk type in there any more
        d.update(copy.deepcopy(full_dict))
        return d

    @classmethod
    def from_dump(cls, dct: Dict, class_kwargs=None) -> "KeyStore":
        super()._from_dump(dct, class_kwargs=class_kwargs)

        return cls(**filtered_for_init(dct, cls))

    @classmethod
    def from_dump_migration(cls, dct: Dict[str, Any]) -> Dict[str, Any]:
        "this class should be overwritten in child classes"
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.0"):
            if "derivation_path" in dct:
                dct["key_origin"] = dct["derivation_path"]
                del dct["derivation_path"]

        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.1"):
            if "derivation_path" in dct:
                dct["network"] = bdk.Network.REGTEST

        # now the version is newest, so it can be deleted from the dict
        if "VERSION" in dct:
            del dct["VERSION"]
        return dct

    def from_other_keystore(self, other_keystore: "KeyStore") -> None:
        for k, v in other_keystore.__dict__.items():
            setattr(self, k, v)

    def is_identical_to(self, spk_provider: SimplePubKeyProvider) -> bool:
        # fill in missing info in keystores

        return (
            self.fingerprint == spk_provider.fingerprint
            and self.xpub == spk_provider.xpub
            and self.key_origin == spk_provider.key_origin
        )
